@@ -55,9 +55,12 @@ generate_fake_email_from_fullname (const gchar *fullname)
 }
 
 static gboolean
-account_miner_job_process_entry (GomAccountMinerJob *job,
+account_miner_job_process_entry (TrackerSparqlConnection *connection,
+                                 GHashTable *previous_resources,
+                                 const gchar *datasource_urn,
                                  GDataDocumentsService *service,
                                  GDataDocumentsEntry *doc_entry,
+                                 GCancellable *cancellable,
                                  GError **error)
 {
   GDataEntry *entry = GDATA_ENTRY (doc_entry);
@@ -96,8 +99,9 @@ account_miner_job_process_entry (GomAccountMinerJob *job,
       identifier = g_strdup_printf ("%s%s", PREFIX_DRIVE, id);
     }
 
-  /* remove from the list of the previous resources */
-  g_hash_table_remove (job->previous_resources, identifier);
+  /* remove from the list of the previous resources, if any */
+  if (previous_resources != NULL)
+    g_hash_table_remove (previous_resources, identifier);
 
   if (GDATA_IS_DOCUMENTS_PRESENTATION (doc_entry))
     class = "nfo:Presentation";
@@ -113,26 +117,26 @@ account_miner_job_process_entry (GomAccountMinerJob *job,
     class = "nfo:DataContainer";
 
   resource = gom_tracker_sparql_connection_ensure_resource
-    (job->connection,
-     job->cancellable, error,
+    (connection,
+     cancellable, error,
      &resource_exists,
-     job->datasource_urn, identifier,
+     datasource_urn, identifier,
      "nfo:RemoteDataObject", class, NULL);
 
   if (*error != NULL)
     goto out;
 
-  gom_tracker_update_datasource (job->connection, job->datasource_urn,
+  gom_tracker_update_datasource (connection, datasource_urn,
                                  resource_exists, identifier, resource,
-                                 job->cancellable, error);
+                                 cancellable, error);
 
   if (*error != NULL)
     goto out;
 
   new_mtime = gdata_entry_get_updated (entry);
-  mtime_changed = gom_tracker_update_mtime (job->connection, new_mtime,
+  mtime_changed = gom_tracker_update_mtime (connection, new_mtime,
                                             resource_exists, identifier, resource,
-                                            job->cancellable, error);
+                                            cancellable, error);
 
   if (*error != NULL)
     goto out;
@@ -148,9 +152,9 @@ account_miner_job_process_entry (GomAccountMinerJob *job,
   alternate_uri = gdata_link_get_uri (alternate);
 
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nie:url", alternate_uri);
 
   if (*error != NULL)
@@ -163,9 +167,9 @@ account_miner_job_process_entry (GomAccountMinerJob *job,
     mimetype_override = "application/pdf";
 
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nie:mimeType", mimetype_override);
 
   if (*error != NULL)
@@ -181,9 +185,9 @@ account_miner_job_process_entry (GomAccountMinerJob *job,
         g_strdup_printf ("gd:collection:%s%s", PREFIX_DRIVE, gdata_link_get_uri (parent));
 
       parent_resource_urn = gom_tracker_sparql_connection_ensure_resource
-        (job->connection, job->cancellable, error,
+        (connection, cancellable, error,
          NULL,
-         job->datasource_urn, parent_resource_id,
+         datasource_urn, parent_resource_id,
          "nfo:RemoteDataObject", "nfo:DataContainer", NULL);
       g_free (parent_resource_id);
 
@@ -191,9 +195,9 @@ account_miner_job_process_entry (GomAccountMinerJob *job,
         goto out;
 
       gom_tracker_sparql_connection_insert_or_replace_triple
-        (job->connection,
-         job->cancellable, error,
-         job->datasource_urn, resource,
+        (connection,
+         cancellable, error,
+         datasource_urn, resource,
          "nie:isPartOf", parent_resource_urn);
       g_free (parent_resource_urn);
 
@@ -213,26 +217,26 @@ account_miner_job_process_entry (GomAccountMinerJob *job,
     }
 
   gom_tracker_sparql_connection_toggle_favorite
-    (job->connection,
-     job->cancellable, error,
+    (connection,
+     cancellable, error,
      resource, starred);
 
   if (*error != NULL)
     goto out;
 
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nie:description", gdata_entry_get_summary (entry));
 
   if (*error != NULL)
     goto out;
 
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nie:title", gdata_entry_get_title (entry));
 
   if (*error != NULL)
@@ -245,8 +249,8 @@ account_miner_job_process_entry (GomAccountMinerJob *job,
 
       author = l->data;
 
-      contact_resource = gom_tracker_utils_ensure_contact_resource (job->connection,
-                                                                    job->cancellable, error,
+      contact_resource = gom_tracker_utils_ensure_contact_resource (connection,
+                                                                    cancellable, error,
                                                                     gdata_author_get_email_address (author),
                                                                     gdata_author_get_name (author));
 
@@ -254,9 +258,9 @@ account_miner_job_process_entry (GomAccountMinerJob *job,
         goto out;
 
       gom_tracker_sparql_connection_insert_or_replace_triple
-        (job->connection,
-         job->cancellable, error,
-         job->datasource_urn, resource,
+        (connection,
+         cancellable, error,
+         datasource_urn, resource,
          "nco:creator", contact_resource);
 
       if (*error != NULL)
@@ -267,7 +271,7 @@ account_miner_job_process_entry (GomAccountMinerJob *job,
 
   access_rules = gdata_access_handler_get_rules (GDATA_ACCESS_HANDLER (entry),
                                                  GDATA_SERVICE (service),
-                                                 job->cancellable,
+                                                 cancellable,
                                                  NULL, NULL, error);
 
   if (*error != NULL)
@@ -290,15 +294,15 @@ account_miner_job_process_entry (GomAccountMinerJob *job,
       if (g_strcmp0 (scope_type, GDATA_ACCESS_SCOPE_DOMAIN) == 0)
         continue;
 
-      contact_resource = gom_tracker_utils_ensure_contact_resource (job->connection,
-                                                                    job->cancellable, error,
+      contact_resource = gom_tracker_utils_ensure_contact_resource (connection,
+                                                                    cancellable, error,
                                                                     scope_value,
                                                                     "");
 
       gom_tracker_sparql_connection_insert_or_replace_triple
-        (job->connection,
-         job->cancellable, error,
-         job->datasource_urn, resource,
+        (connection,
+         cancellable, error,
+         datasource_urn, resource,
          "nco:contributor", contact_resource);
 
       g_free (contact_resource);
@@ -309,9 +313,9 @@ account_miner_job_process_entry (GomAccountMinerJob *job,
 
   date = gom_iso8601_from_timestamp (gdata_entry_get_published (entry));
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nie:contentCreated", date);
   g_free (date);
 
@@ -331,10 +335,13 @@ account_miner_job_process_entry (GomAccountMinerJob *job,
   return TRUE;
 }
 
-static gboolean
-account_miner_job_process_photo (GomAccountMinerJob *job,
+static gchar *
+account_miner_job_process_photo (TrackerSparqlConnection *connection,
+                                 GHashTable *previous_resources,
+                                 const gchar *datasource_urn,
                                  GDataPicasaWebFile *photo,
                                  const gchar *parent_resource_urn,
+                                 GCancellable *cancellable,
                                  GError **error)
 {
   GList *l, *media_contents;
@@ -385,22 +392,23 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
 
   identifier = g_strdup_printf ("%s%s", PREFIX_PICASAWEB, id);
 
-  /* remove from the list of the previous resources */
-  g_hash_table_remove (job->previous_resources, identifier);
+  /* remove from the list of the previous resources, if any */
+  if (previous_resources != NULL)
+    g_hash_table_remove (previous_resources, identifier);
 
   resource = gom_tracker_sparql_connection_ensure_resource
-    (job->connection,
-     job->cancellable, error,
+    (connection,
+     cancellable, error,
      &resource_exists,
-     job->datasource_urn, identifier,
+     datasource_urn, identifier,
      "nfo:RemoteDataObject", "nmm:Photo", NULL);
 
   if (*error != NULL)
     goto out;
 
-  gom_tracker_update_datasource (job->connection, job->datasource_urn,
+  gom_tracker_update_datasource (connection, datasource_urn,
                                  resource_exists, identifier, resource,
-                                 job->cancellable, error);
+                                 cancellable, error);
   if (*error != NULL)
     goto out;
 
@@ -408,9 +416,9 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
    * been modified since our last run
    */
   new_mtime = gdata_entry_get_updated (GDATA_ENTRY (photo));
-  mtime_changed = gom_tracker_update_mtime (job->connection, new_mtime,
+  mtime_changed = gom_tracker_update_mtime (connection, new_mtime,
                                             resource_exists, identifier, resource,
-                                            job->cancellable, error);
+                                            cancellable, error);
 
   if (*error != NULL)
     goto out;
@@ -425,9 +433,9 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
   alternate = gdata_entry_look_up_link (GDATA_ENTRY (photo), GDATA_LINK_ALTERNATE);
   alternate_uri = gdata_link_get_uri (alternate);
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nie:url", alternate_uri);
 
   if (*error != NULL)
@@ -435,28 +443,31 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
 
   summary = gdata_entry_get_summary ((GDATA_ENTRY (photo)));
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nie:description", summary);
 
   if (*error != NULL)
     goto out;
 
-  gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
-     "nie:isPartOf", parent_resource_urn);
+  if (parent_resource_urn != NULL)
+    {
+      gom_tracker_sparql_connection_insert_or_replace_triple
+        (connection,
+         cancellable, error,
+         datasource_urn, resource,
+         "nie:isPartOf", parent_resource_urn);
 
-  if (*error != NULL)
-    goto out;
+      if (*error != NULL)
+        goto out;
+    }
 
   mime = gdata_media_content_get_content_type (GDATA_MEDIA_CONTENT (media_contents->data));
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nie:mimeType", mime);
 
   if (*error != NULL)
@@ -464,9 +475,9 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
 
   title = gdata_entry_get_title ((GDATA_ENTRY (photo)));
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nie:title", title);
 
   if (*error != NULL)
@@ -475,8 +486,8 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
   credit = gdata_picasaweb_file_get_credit (photo);
   email = generate_fake_email_from_fullname (credit);
   contact_resource = gom_tracker_utils_ensure_contact_resource
-    (job->connection,
-     job->cancellable, error,
+    (connection,
+     cancellable, error,
      email, credit);
   g_free (email);
 
@@ -484,9 +495,9 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
     goto out;
 
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nco:creator", contact_resource);
 
   g_free (contact_resource);
@@ -495,9 +506,9 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
 
   exposure = g_strdup_printf ("%f", gdata_picasaweb_file_get_exposure (photo));
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nmm:exposureTime", exposure);
   g_free (exposure);
 
@@ -506,9 +517,9 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
 
   focal_length = g_strdup_printf ("%f", gdata_picasaweb_file_get_focal_length (photo));
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nmm:focalLength", focal_length);
   g_free (focal_length);
 
@@ -517,9 +528,9 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
 
   fstop = g_strdup_printf ("%f", gdata_picasaweb_file_get_fstop (photo));
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nmm:fnumber", fstop);
   g_free (fstop);
 
@@ -528,9 +539,9 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
 
   iso = g_strdup_printf ("%ld", (glong) gdata_picasaweb_file_get_iso (photo));
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nmm:isoSpeed", iso);
   g_free (iso);
 
@@ -539,9 +550,9 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
 
   flash = gdata_picasaweb_file_get_flash (photo);
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nmm:flash", flash ? flash_on : flash_off);
 
   if (*error != NULL)
@@ -552,8 +563,8 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
 
   if (make != NULL || model != NULL)
     {
-      equipment_resource = gom_tracker_utils_ensure_equipment_resource (job->connection,
-                                                                        job->cancellable,
+      equipment_resource = gom_tracker_utils_ensure_equipment_resource (connection,
+                                                                        cancellable,
                                                                         error,
                                                                         make,
                                                                         model);
@@ -562,9 +573,9 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
         goto out;
 
       gom_tracker_sparql_connection_insert_or_replace_triple
-        (job->connection,
-         job->cancellable, error,
-         job->datasource_urn, resource,
+        (connection,
+         cancellable, error,
+         datasource_urn, resource,
          "nfo:equipment", equipment_resource);
 
       if (*error != NULL)
@@ -573,9 +584,9 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
 
   width = g_strdup_printf ("%u", gdata_picasaweb_file_get_width (photo));
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nfo:width", width);
   g_free (width);
 
@@ -584,9 +595,9 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
 
   height = g_strdup_printf ("%u", gdata_picasaweb_file_get_height (photo));
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nfo:height", height);
   g_free (height);
 
@@ -596,9 +607,9 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
   timestamp = gdata_picasaweb_file_get_timestamp (photo);
   date = gom_iso8601_from_timestamp (timestamp / 1000);
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nie:contentCreated", date);
   g_free (date);
 
@@ -606,20 +617,22 @@ account_miner_job_process_photo (GomAccountMinerJob *job,
     goto out;
 
  out:
-  g_free (resource);
   g_free (identifier);
   g_free (equipment_resource);
 
   if (*error != NULL)
-    return FALSE;
+    return NULL;
 
-  return TRUE;
+  return resource;
 }
 
 static gboolean
-account_miner_job_process_album (GomAccountMinerJob *job,
+account_miner_job_process_album (TrackerSparqlConnection *connection,
+                                 GHashTable *previous_resources,
+                                 const gchar *datasource_urn,
                                  GDataPicasaWebService *service,
                                  GDataPicasaWebAlbum *album,
+                                 GCancellable *cancellable,
                                  GError **error)
 {
   GDataFeed *feed = NULL;
@@ -644,14 +657,15 @@ account_miner_job_process_album (GomAccountMinerJob *job,
   album_id = gdata_entry_get_id (GDATA_ENTRY (album));
   identifier = g_strdup_printf ("photos:collection:%s%s", PREFIX_PICASAWEB, album_id);
 
-  /* remove from the list of the previous resources */
-  g_hash_table_remove (job->previous_resources, identifier);
+  /* remove from the list of the previous resources, if any */
+  if (previous_resources != NULL)
+    g_hash_table_remove (previous_resources, identifier);
 
   resource = gom_tracker_sparql_connection_ensure_resource
-    (job->connection,
-     job->cancellable, error,
+    (connection,
+     cancellable, error,
      &resource_exists,
-     job->datasource_urn, identifier,
+     datasource_urn, identifier,
      "nfo:RemoteDataObject", "nfo:DataContainer",
      NULL);
 
@@ -659,9 +673,9 @@ account_miner_job_process_album (GomAccountMinerJob *job,
     goto out;
 
   gom_tracker_update_datasource
-    (job->connection, job->datasource_urn,
+    (connection, datasource_urn,
      resource_exists, identifier, resource,
-     job->cancellable, error);
+     cancellable, error);
 
   if (*error != NULL)
     goto out;
@@ -670,9 +684,9 @@ account_miner_job_process_album (GomAccountMinerJob *job,
    * been modified since our last run
    */
   new_mtime = gdata_entry_get_updated (GDATA_ENTRY (album));
-  mtime_changed = gom_tracker_update_mtime (job->connection, new_mtime,
+  mtime_changed = gom_tracker_update_mtime (connection, new_mtime,
                                             resource_exists, identifier, resource,
-                                            job->cancellable, error);
+                                            cancellable, error);
 
   if (*error != NULL)
     goto out;
@@ -687,9 +701,9 @@ account_miner_job_process_album (GomAccountMinerJob *job,
   alternate = gdata_entry_look_up_link (GDATA_ENTRY (album), GDATA_LINK_ALTERNATE);
   alternate_uri = gdata_link_get_uri (alternate);
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nie:url", alternate_uri);
 
   if (*error != NULL)
@@ -697,9 +711,9 @@ account_miner_job_process_album (GomAccountMinerJob *job,
 
   summary = gdata_entry_get_summary ((GDATA_ENTRY (album)));
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nie:description", summary);
 
   if (*error != NULL)
@@ -707,9 +721,9 @@ account_miner_job_process_album (GomAccountMinerJob *job,
 
   title = gdata_entry_get_title ((GDATA_ENTRY (album)));
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nie:title", title);
 
   if (*error != NULL)
@@ -718,8 +732,8 @@ account_miner_job_process_album (GomAccountMinerJob *job,
   nickname = gdata_picasaweb_album_get_nickname (album);
   email = generate_fake_email_from_fullname (nickname);
   contact_resource = gom_tracker_utils_ensure_contact_resource
-    (job->connection,
-     job->cancellable, error,
+    (connection,
+     cancellable, error,
      email, nickname);
   g_free (email);
 
@@ -727,9 +741,9 @@ account_miner_job_process_album (GomAccountMinerJob *job,
     goto out;
 
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nco:creator", contact_resource);
   g_free (contact_resource);
 
@@ -739,9 +753,9 @@ account_miner_job_process_album (GomAccountMinerJob *job,
   timestamp = gdata_picasaweb_album_get_timestamp (album);
   date = gom_iso8601_from_timestamp (timestamp / 1000);
   gom_tracker_sparql_connection_insert_or_replace_triple
-    (job->connection,
-     job->cancellable, error,
-     job->datasource_urn, resource,
+    (connection,
+     cancellable, error,
+     datasource_urn, resource,
      "nie:contentCreated", date);
   g_free (date);
 
@@ -752,7 +766,7 @@ account_miner_job_process_album (GomAccountMinerJob *job,
   query = gdata_picasaweb_query_new (NULL);
   gdata_picasaweb_query_set_image_size (query, "d");
   feed = gdata_picasaweb_service_query_files (service, album, GDATA_QUERY (query),
-                                              job->cancellable, NULL, NULL, error);
+                                              cancellable, NULL, NULL, error);
 
   g_object_unref (query);
 
@@ -763,8 +777,15 @@ account_miner_job_process_album (GomAccountMinerJob *job,
   for (l = photos; l != NULL; l = l->next)
     {
       GDataPicasaWebFile *file = GDATA_PICASAWEB_FILE (l->data);
+      gchar *photo_resource_urn = NULL;
 
-      account_miner_job_process_photo (job, file, resource, error);
+      photo_resource_urn = account_miner_job_process_photo (connection,
+                                                            previous_resources,
+                                                            datasource_urn,
+                                                            file,
+                                                            resource,
+                                                            cancellable,
+                                                            error);
 
       if (*error != NULL)
         {
@@ -774,6 +795,8 @@ account_miner_job_process_album (GomAccountMinerJob *job,
           g_warning ("Unable to process photo %s: %s", photo_id, (*error)->message);
           g_clear_error (error);
         }
+
+      g_free (photo_resource_urn);
     }
 
  out:
@@ -787,9 +810,117 @@ account_miner_job_process_album (GomAccountMinerJob *job,
   return TRUE;
 }
 
+static void
+insert_shared_content_photos (TrackerSparqlConnection *connection,
+                              const gchar *datasource_urn,
+                              const gchar *shared_id,
+                              const gchar *source_urn,
+                              GDataPicasaWebService *service,
+                              GCancellable *cancellable,
+                              GError **error)
+{
+  GError *local_error;
+  GDataAuthorizationDomain *authorization_domain;
+  GDataEntry *entry = NULL;
+  GDataPicasaWebFile *file;
+  GDataPicasaWebQuery *query = NULL;
+  gchar *photo_resource_urn = NULL;
+
+  authorization_domain = gdata_picasaweb_service_get_primary_authorization_domain ();
+
+  query = gdata_picasaweb_query_new (NULL);
+  gdata_picasaweb_query_set_image_size (query, "d");
+
+  local_error = NULL;
+  entry = gdata_service_query_single_entry (GDATA_SERVICE (service),
+                                            authorization_domain,
+                                            shared_id,
+                                            GDATA_QUERY (query),
+                                            GDATA_TYPE_PICASAWEB_FILE,
+                                            cancellable,
+                                            &local_error);
+  if (local_error != NULL)
+    {
+      g_propagate_error (error, local_error);
+      goto out;
+    }
+
+  file = GDATA_PICASAWEB_FILE (entry);
+
+  local_error = NULL;
+  photo_resource_urn = account_miner_job_process_photo (connection,
+                                                        NULL,
+                                                        datasource_urn,
+                                                        file,
+                                                        NULL,
+                                                        cancellable,
+                                                        &local_error);
+  if (local_error != NULL)
+    {
+      g_propagate_error (error, local_error);
+      goto out;
+    }
+
+  local_error = NULL;
+  if (!gom_tracker_sparql_connection_insert_or_replace_triple (connection,
+                                                               cancellable,
+                                                               &local_error,
+                                                               datasource_urn,
+                                                               source_urn,
+                                                               "nie:relatedTo",
+                                                               photo_resource_urn))
+    {
+      g_propagate_error (error, local_error);
+      goto out;
+    }
+
+  local_error = NULL;
+  if (!gom_tracker_sparql_connection_insert_or_replace_triple (connection,
+                                                               cancellable,
+                                                               &local_error,
+                                                               datasource_urn,
+                                                               photo_resource_urn,
+                                                               "nie:links",
+                                                               source_urn))
+    {
+      g_propagate_error (error, local_error);
+      goto out;
+    }
+
+ out:
+  g_clear_object (&entry);
+  g_clear_object (&query);
+  g_free (photo_resource_urn);
+}
+
+static void
+insert_shared_content (GomMiner *miner,
+                       gpointer service,
+                       TrackerSparqlConnection *connection,
+                       const gchar *datasource_urn,
+                       const gchar *shared_id,
+                       const gchar *shared_type,
+                       const gchar *source_urn,
+                       GCancellable *cancellable,
+                       GError **error)
+{
+  if (g_strcmp0 (shared_type, "photos") == 0)
+    insert_shared_content_photos (connection,
+                                  datasource_urn,
+                                  shared_id,
+                                  source_urn,
+                                  GDATA_PICASAWEB_SERVICE (service),
+                                  cancellable,
+                                  error);
+}
+
 static void
 query_gdata_documents (GomAccountMinerJob *job,
+                       TrackerSparqlConnection *connection,
+                       GHashTable *previous_resources,
+                       const gchar *datasource_urn,
                        GDataDocumentsService *service,
+                       GCancellable *cancellable,
                        GError **error)
 {
   GDataDocumentsQuery *query = NULL;
@@ -807,7 +938,7 @@ query_gdata_documents (GomAccountMinerJob *job,
       local_error = NULL;
       feed = gdata_documents_service_query_documents
         (service, query,
-         job->cancellable, NULL, NULL, &local_error);
+         cancellable, NULL, NULL, &local_error);
       if (local_error != NULL)
         {
           if (succeeded_once)
@@ -832,7 +963,13 @@ query_gdata_documents (GomAccountMinerJob *job,
       for (l = entries; l != NULL; l = l->next)
         {
           local_error = NULL;
-          account_miner_job_process_entry (job, service, l->data, &local_error);
+          account_miner_job_process_entry (connection,
+                                           previous_resources,
+                                           datasource_urn,
+                                           service,
+                                           l->data,
+                                           cancellable,
+                                           &local_error);
 
           if (local_error != NULL)
             {
@@ -852,13 +989,17 @@ query_gdata_documents (GomAccountMinerJob *job,
 
 static void
 query_gdata_photos (GomAccountMinerJob *job,
+                    TrackerSparqlConnection *connection,
+                    GHashTable *previous_resources,
+                    const gchar *datasource_urn,
                     GDataPicasaWebService *service,
+                    GCancellable *cancellable,
                     GError **error)
 {
   GDataFeed *feed;
   GList *albums, *l;
 
-  feed = gdata_picasaweb_service_query_all_albums (service, NULL, NULL, job->cancellable, NULL, NULL, error);
+  feed = gdata_picasaweb_service_query_all_albums (service, NULL, NULL, cancellable, NULL, NULL, error);
 
   if (feed == NULL)
     return;
@@ -868,7 +1009,13 @@ query_gdata_photos (GomAccountMinerJob *job,
     {
       GDataPicasaWebAlbum *album = GDATA_PICASAWEB_ALBUM (l->data);
 
-      account_miner_job_process_album (job, service, album, error);
+      account_miner_job_process_album (connection,
+                                       previous_resources,
+                                       datasource_urn,
+                                       service,
+                                       album,
+                                       cancellable,
+                                       error);
 
       if (*error != NULL)
         {
@@ -885,17 +1032,51 @@ query_gdata_photos (GomAccountMinerJob *job,
 
 static void
 query_gdata (GomAccountMinerJob *job,
+             TrackerSparqlConnection *connection,
+             GHashTable *previous_resources,
+             const gchar *datasource_urn,
+             GCancellable *cancellable,
              GError **error)
 {
   gpointer service;
 
   service = g_hash_table_lookup (job->services, "documents");
   if (service != NULL)
-    query_gdata_documents (job, GDATA_DOCUMENTS_SERVICE (service), error);
+    query_gdata_documents (job,
+                           connection,
+                           previous_resources,
+                           datasource_urn,
+                           GDATA_DOCUMENTS_SERVICE (service),
+                           cancellable,
+                           error);
 
   service = g_hash_table_lookup (job->services, "photos");
   if (service != NULL)
-    query_gdata_photos (job, GDATA_PICASAWEB_SERVICE (service), error);
+    query_gdata_photos (job,
+                        connection,
+                        previous_resources,
+                        datasource_urn,
+                        GDATA_PICASAWEB_SERVICE (service),
+                        cancellable,
+                        error);
+}
+
+static gpointer
+create_service (GomMiner *miner, GoaObject *object, const gchar *type)
+{
+  GDataGoaAuthorizer *authorizer;
+  gpointer service = NULL;
+
+  authorizer = gdata_goa_authorizer_new (object);
+
+  if (g_strcmp0 (type, "documents") == 0)
+    service = gdata_documents_service_new (GDATA_AUTHORIZER (authorizer));
+
+  if (g_strcmp0 (type, "photos") == 0)
+    service = gdata_picasaweb_service_new (GDATA_AUTHORIZER (authorizer));
+
+  g_object_unref (authorizer);
+  return service;
 }
 
 static GHashTable *
@@ -931,6 +1112,12 @@ create_services (GomMiner *self,
   return services;
 }
 
+static void
+destroy_service (GomMiner *miner, gpointer service)
+{
+  g_object_unref (service);
+}
+
 static void
 gom_gdata_miner_init (GomGDataMiner *miner)
 {
@@ -945,6 +1132,9 @@ gom_gdata_miner_class_init (GomGDataMinerClass *klass)
   miner_class->miner_identifier = MINER_IDENTIFIER;
   miner_class->version = 5;
 
+  miner_class->create_service = create_service;
   miner_class->create_services = create_services;
+  miner_class->destroy_service = destroy_service;
+  miner_class->insert_shared_content = insert_shared_content;
   miner_class->query = query_gdata;
 }
