@@ -35,27 +35,42 @@ struct _GomMinerPrivate {
   GError *client_error;
 
   TrackerSparqlConnection *connection;
-
-  GCancellable *cancellable;
-  GSimpleAsyncResult *result;
-
-  GList *pending_jobs;
+  GError *connection_error;
 
   gchar *display_name;
   gchar **index_types;
 };
 
+typedef struct {
+  GomMiner *self;
+  GList *content_objects;
+  GList *acc_objects;
+  GList *old_datasources;
+  GList *pending_jobs;
+} CleanupJob;
+
+typedef struct {
+  GomMiner *self;
+  gchar *account_id;
+  gchar *shared_id;
+  gchar *shared_type;
+  gchar *source_urn;
+  gpointer service;
+} InsertSharedContentData;
+
+static GThreadPool *cleanup_pool;
+
+static void cleanup_job (gpointer data, gpointer user_data);
+
 static void
 gom_account_miner_job_free (GomAccountMinerJob *job)
 {
-  if (job->miner_cancellable_id != 0)
-    g_cancellable_disconnect (job->miner->priv->cancellable,
-                              job->miner_cancellable_id);
-
   g_hash_table_unref (job->services);
   g_clear_object (&job->miner);
   g_clear_object (&job->account);
-  g_clear_object (&job->async_result);
+  g_clear_object (&job->connection);
+  g_clear_object (&job->task);
+  g_clear_object (&job->parent_task);
 
   g_free (job->datasource_urn);
   g_free (job->root_element_urn);
@@ -65,26 +80,53 @@ gom_account_miner_job_free (GomAccountMinerJob *job)
   g_slice_free (GomAccountMinerJob, job);
 }
 
+static void
+gom_insert_shared_content_data_free (InsertSharedContentData *data)
+{
+  GOM_MINER_GET_CLASS (data->self)->destroy_service (data->self, data->service);
+
+  g_object_unref (data->self);
+  g_free (data->account_id);
+  g_free (data->shared_id);
+  g_free (data->shared_type);
+  g_free (data->source_urn);
+
+  g_slice_free (InsertSharedContentData, data);
+}
+
+static InsertSharedContentData *
+gom_insert_shared_content_data_new (GomMiner *self,
+                                    const gchar *account_id,
+                                    const gchar *shared_id,
+                                    const gchar *shared_type,
+                                    const gchar *source_urn,
+                                    gpointer service)
+{
+  InsertSharedContentData *retval;
+
+  retval = g_slice_new0 (InsertSharedContentData);
+  retval->self = g_object_ref (self);
+  retval->account_id = g_strdup (account_id);
+  retval->shared_id = g_strdup (shared_id);
+  retval->shared_type = g_strdup (shared_type);
+  retval->source_urn = g_strdup (source_urn);
+  retval->service = service;
+
+  return retval;
+}
+
 static void
 gom_miner_dispose (GObject *object)
 {
   GomMiner *self = GOM_MINER (object);
 
-  if (self->priv->pending_jobs != NULL)
-    {
-      g_list_free_full (self->priv->pending_jobs,
-                        (GDestroyNotify) gom_account_miner_job_free);
-      self->priv->pending_jobs = NULL;
-    }
-
   g_clear_object (&self->priv->client);
   g_clear_object (&self->priv->connection);
-  g_clear_object (&self->priv->cancellable);
-  g_clear_object (&self->priv->result);
 
   g_free (self->priv->display_name);
   g_strfreev (self->priv->index_types);
   g_clear_error (&self->priv->client_error);
+  g_clear_error (&self->priv->connection_error);
 
   G_OBJECT_CLASS (gom_miner_parent_class)->dispose (object);
 }
@@ -141,8 +183,18 @@ gom_miner_constructed (GObject *obj)
 static void
 gom_miner_init (GomMiner *self)
 {
+  GomMinerClass *klass = GOM_MINER_GET_CLASS (self);
+
   self->priv = G_TYPE_INSTANCE_GET_PRIVATE (self, GOM_TYPE_MINER, GomMinerPrivate);
   self->priv->display_name = g_strdup ("");
+
+  self->priv->connection = tracker_sparql_connection_get (NULL, &self->priv->connection_error);
+  if (self->priv->connection_error != NULL)
+    {
+      g_critical ("Unable to create TrackerSparqlConnection: %s - indexing for %s will not work",
+                  self->priv->connection_error->message,
+                  klass->goa_provider_type);
+    }
 }
 
 static void
@@ -153,32 +205,34 @@ gom_miner_class_init (GomMinerClass *klass)
   oclass->constructed = gom_miner_constructed;
   oclass->dispose = gom_miner_dispose;
 
+  cleanup_pool = g_thread_pool_new (cleanup_job, NULL, 1, FALSE, NULL);
+
   g_type_class_add_private (klass, sizeof (GomMinerPrivate));
 }
 
 static void
-gom_miner_complete_error (GomMiner *self,
-                          GError *error)
+gom_miner_check_pending_jobs (GTask *task)
 {
-  g_assert (self->priv->result != NULL);
+  CleanupJob *cleanup_job;
 
-  g_simple_async_result_take_error (self->priv->result, error);
-  g_simple_async_result_complete_in_idle (self->priv->result);
+  cleanup_job = (CleanupJob *) g_task_get_task_data (task);
+
+  if (g_list_length (cleanup_job->pending_jobs) > 0)
+    return;
+
+  g_task_return_boolean (task, TRUE);
+  g_slice_free (CleanupJob, cleanup_job);
 }
 
 static void
-gom_miner_check_pending_jobs (GomMiner *self)
-{
-  if (g_list_length (self->priv->pending_jobs) == 0)
-    g_simple_async_result_complete_in_idle (self->priv->result);
-}
-
-static void
-gom_account_miner_job_ensure_datasource (GomAccountMinerJob *job,
-                                         GError **error)
+gom_miner_ensure_datasource (GomMiner *self,
+                             const gchar *datasource_urn,
+                             const gchar *root_element_urn,
+                             GCancellable *cancellable,
+                             GError **error)
 {
   GString *datasource_insert;
-  GomMinerClass *klass = GOM_MINER_GET_CLASS (job->miner);
+  GomMinerClass *klass = GOM_MINER_GET_CLASS (self);
 
   datasource_insert = g_string_new (NULL);
   g_string_append_printf (datasource_insert,
@@ -186,14 +240,14 @@ gom_account_miner_job_ensure_datasource (GomAccountMinerJob *job,
                           "  <%s> a nie:DataSource ; nao:identifier \"%s\" . "
                           "  <%s> a nie:InformationElement ; nie:rootElementOf <%s> ; nie:version \"%d\""
                           "}",
-                          job->datasource_urn,
-                          job->datasource_urn, klass->miner_identifier,
-                          job->root_element_urn, job->datasource_urn, klass->version);
+                          datasource_urn,
+                          datasource_urn, klass->miner_identifier,
+                          root_element_urn, datasource_urn, klass->version);
 
-  tracker_sparql_connection_update (job->connection,
+  tracker_sparql_connection_update (self->priv->connection,
                                     datasource_insert->str,
                                     G_PRIORITY_DEFAULT,
-                                    job->cancellable,
+                                    cancellable,
                                     error);
 
   g_string_free (datasource_insert, TRUE);
@@ -203,9 +257,12 @@ static void
 gom_account_miner_job_query_existing (GomAccountMinerJob *job,
                                       GError **error)
 {
+  GCancellable *cancellable;
   GString *select;
   TrackerSparqlCursor *cursor;
 
+  cancellable = g_task_get_cancellable (job->task);
+
   select = g_string_new (NULL);
   g_string_append_printf (select,
                           "SELECT ?urn nao:identifier(?urn) WHERE { ?urn nie:dataSource <%s> }",
@@ -213,14 +270,14 @@ gom_account_miner_job_query_existing (GomAccountMinerJob *job,
 
   cursor = tracker_sparql_connection_query (job->connection,
                                             select->str,
-                                            job->cancellable,
+                                            cancellable,
                                             error);
   g_string_free (select, TRUE);
 
   if (cursor == NULL)
     return;
 
-  while (tracker_sparql_cursor_next (cursor, job->cancellable, error))
+  while (tracker_sparql_cursor_next (cursor, cancellable, error))
     {
       g_hash_table_insert (job->previous_resources,
                            g_strdup (tracker_sparql_cursor_get_string (cursor, 1, NULL)),
@@ -245,8 +302,11 @@ static void
 gom_account_miner_job_cleanup_previous (GomAccountMinerJob *job,
                                         GError **error)
 {
+  GCancellable *cancellable;
   GString *delete;
 
+  cancellable = g_task_get_cancellable (job->task);
+
   delete = g_string_new (NULL);
   g_string_append (delete, "DELETE { ");
 
@@ -262,7 +322,7 @@ gom_account_miner_job_cleanup_previous (GomAccountMinerJob *job,
   tracker_sparql_connection_update (job->connection,
                                     delete->str,
                                     G_PRIORITY_DEFAULT,
-                                    job->cancellable,
+                                    cancellable,
                                     error);
 
   g_string_free (delete, TRUE);
@@ -273,19 +333,22 @@ gom_account_miner_job_query (GomAccountMinerJob *job,
                              GError **error)
 {
   GomMinerClass *miner_class = GOM_MINER_GET_CLASS (job->miner);
+  GCancellable *cancellable;
 
-  miner_class->query (job, error);
+  cancellable = g_task_get_cancellable (job->task);
+  miner_class->query (job, job->connection, job->previous_resources, job->datasource_urn, cancellable, error);
 }
 
-static gboolean
-gom_account_miner_job (GIOSchedulerJob *sched_job,
-                       GCancellable *cancellable,
-                       gpointer user_data)
+static void
+gom_account_miner_job (GTask *task,
+                       gpointer source_object,
+                       gpointer task_data,
+                       GCancellable *cancellable)
 {
-  GomAccountMinerJob *job = user_data;
+  GomAccountMinerJob *job = task_data;
   GError *error = NULL;
 
-  gom_account_miner_job_ensure_datasource (job, &error);
+  gom_miner_ensure_datasource (job->miner, job->datasource_urn, job->root_element_urn, cancellable, &error);
 
   if (error != NULL)
     goto out;
@@ -307,11 +370,9 @@ gom_account_miner_job (GIOSchedulerJob *sched_job,
 
  out:
   if (error != NULL)
-    g_simple_async_result_take_error (job->async_result, error);
-
-  g_simple_async_result_complete_in_idle (job->async_result);
-
-  return FALSE;
+    g_task_return_error (job->task, error);
+  else
+    g_task_return_boolean (job->task, TRUE);
 }
 
 static void
@@ -319,45 +380,36 @@ gom_account_miner_job_process_async (GomAccountMinerJob *job,
                                      GAsyncReadyCallback callback,
                                      gpointer user_data)
 {
-  g_assert (job->async_result == NULL);
+  GCancellable *cancellable;
 
-  job->async_result = g_simple_async_result_new (NULL, callback, user_data,
-                                                 gom_account_miner_job_process_async);
-  g_simple_async_result_set_op_res_gpointer (job->async_result, job, NULL);
+  g_assert (job->task == NULL);
 
-  g_io_scheduler_push_job (gom_account_miner_job, job, NULL,
-                           G_PRIORITY_DEFAULT,
-                           job->cancellable);
+  cancellable = g_task_get_cancellable (job->parent_task);
+
+  job->task = g_task_new (NULL, cancellable, callback, user_data);
+  g_task_set_source_tag (job->task, gom_account_miner_job_process_async);
+  g_task_set_task_data (job->task, job, NULL);
+  g_task_run_in_thread (job->task, gom_account_miner_job);
 }
 
 static gboolean
 gom_account_miner_job_process_finish (GAsyncResult *res,
                                       GError **error)
 {
-  GSimpleAsyncResult *simple_res = G_SIMPLE_ASYNC_RESULT (res);
+  GTask *task;
 
-  g_assert (g_simple_async_result_is_valid (res, NULL,
-                                            gom_account_miner_job_process_async));
+  g_assert (g_task_is_valid (res, NULL));
+  task = G_TASK (res);
 
-  if (g_simple_async_result_propagate_error (simple_res, error))
-    return FALSE;
+  g_assert (g_task_get_source_tag (task) == gom_account_miner_job_process_async);
 
-  return TRUE;
-}
-
-static void
-miner_cancellable_cancelled_cb (GCancellable *cancellable,
-                                gpointer user_data)
-{
-  GomAccountMinerJob *job = user_data;
-
-  /* forward the cancel signal to the ongoing job */
-  g_cancellable_cancel (job->cancellable);
+  return g_task_propagate_boolean (task, error);
 }
 
 static GomAccountMinerJob *
 gom_account_miner_job_new (GomMiner *self,
-                           GoaObject *object)
+                           GoaObject *object,
+                           GTask *parent_task)
 {
   GomAccountMinerJob *retval;
   GoaAccount *account;
@@ -368,19 +420,13 @@ gom_account_miner_job_new (GomMiner *self,
 
   retval = g_slice_new0 (GomAccountMinerJob);
   retval->miner = g_object_ref (self);
-  retval->cancellable = g_cancellable_new ();
+  retval->parent_task = g_object_ref (parent_task);
   retval->account = account;
-  retval->connection = self->priv->connection;
+  retval->connection = g_object_ref (self->priv->connection);
   retval->previous_resources =
     g_hash_table_new_full (g_str_hash, g_str_equal,
                            (GDestroyNotify) g_free, (GDestroyNotify) g_free);
 
-  if (self->priv->cancellable != NULL)
-      retval->miner_cancellable_id =
-        g_cancellable_connect (self->priv->cancellable,
-                               G_CALLBACK (miner_cancellable_cancelled_cb),
-                               retval, NULL);
-
   retval->services = miner_class->create_services (self, object);
   retval->datasource_urn = g_strdup_printf ("gd:goa-account:%s",
                                             goa_account_get_id (retval->account));
@@ -395,59 +441,63 @@ miner_job_process_ready_cb (GObject *source,
                             GAsyncResult *res,
                             gpointer user_data)
 {
-  GomAccountMinerJob *job = user_data;
-  GomMiner *self = job->miner;
+  CleanupJob *cleanup_job;
+  GomAccountMinerJob *account_miner_job = user_data;
+  GomMiner *self = account_miner_job->miner;
   GError *error = NULL;
 
+  cleanup_job = (CleanupJob *) g_task_get_task_data (account_miner_job->parent_task);
+
   gom_account_miner_job_process_finish (res, &error);
 
   if (error != NULL)
     {
       g_printerr ("Error while refreshing account %s: %s",
-                  goa_account_get_id (job->account), error->message);
+                  goa_account_get_id (account_miner_job->account), error->message);
 
       g_error_free (error);
     }
 
-  self->priv->pending_jobs = g_list_remove (self->priv->pending_jobs,
-                                            job);
-  gom_account_miner_job_free (job);
+  cleanup_job->pending_jobs = g_list_remove (cleanup_job->pending_jobs,
+                                             account_miner_job);
 
-  gom_miner_check_pending_jobs (self);
+  gom_miner_check_pending_jobs (account_miner_job->parent_task);
+  gom_account_miner_job_free (account_miner_job);
 }
 
 static void
 gom_miner_setup_account (GomMiner *self,
-                         GoaObject *object)
+                         GoaObject *object,
+                         GTask *task)
 {
-  GomAccountMinerJob *job;
+  CleanupJob *cleanup_job;
+  GomAccountMinerJob *account_miner_job;
 
-  job = gom_account_miner_job_new (self, object);
-  self->priv->pending_jobs = g_list_prepend (self->priv->pending_jobs, job);
+  cleanup_job = (CleanupJob *) g_task_get_task_data (task);
 
-  gom_account_miner_job_process_async (job, miner_job_process_ready_cb, job);
+  account_miner_job = gom_account_miner_job_new (self, object, task);
+  cleanup_job->pending_jobs = g_list_prepend (cleanup_job->pending_jobs, account_miner_job);
+
+  gom_account_miner_job_process_async (account_miner_job, miner_job_process_ready_cb, account_miner_job);
 }
 
-typedef struct {
-  GomMiner *self;
-  GList *content_objects;
-  GList *acc_objects;
-  GList *old_datasources;
-} CleanupJob;
-
 static gboolean
 cleanup_old_accounts_done (gpointer data)
 {
-  CleanupJob *job = data;
+  GTask *task = G_TASK (data);
+  CleanupJob *job;
   GList *l;
   GoaObject *object;
-  GomMiner *self = job->self;
+  GomMiner *self;
+
+  job = (CleanupJob *) g_task_get_task_data (task);
+  self = job->self;
 
   /* now setup all the current accounts */
   for (l = job->content_objects; l != NULL; l = l->next)
     {
       object = l->data;
-      gom_miner_setup_account (self, object);
+      gom_miner_setup_account (self, object, task);
 
       g_object_unref (object);
     }
@@ -470,16 +520,15 @@ cleanup_old_accounts_done (gpointer data)
       job->old_datasources = NULL;
     }
 
-  gom_miner_check_pending_jobs (self);
+  gom_miner_check_pending_jobs (task);
 
   g_clear_object (&job->self);
-  g_slice_free (CleanupJob, job);
 
   return FALSE;
 }
 
 static void
-cleanup_job_do_cleanup (CleanupJob *job)
+cleanup_job_do_cleanup (CleanupJob *job, GCancellable *cancellable)
 {
   GomMiner *self = job->self;
   GList *l;
@@ -510,7 +559,7 @@ cleanup_job_do_cleanup (CleanupJob *job)
   tracker_sparql_connection_update (self->priv->connection,
                                     update->str,
                                     G_PRIORITY_DEFAULT,
-                                    self->priv->cancellable,
+                                    cancellable,
                                     &error);
   g_string_free (update, TRUE);
 
@@ -543,20 +592,27 @@ cleanup_datasource_compare (gconstpointer a,
   return res;
 }
 
-static gboolean
-cleanup_job (GIOSchedulerJob *sched_job,
-             GCancellable *cancellable,
+static void
+cleanup_job (gpointer data,
              gpointer user_data)
 {
+  GCancellable *cancellable;
+  GSource *source;
   GString *select;
+  GTask *task = G_TASK (data);
   GError *error = NULL;
   TrackerSparqlCursor *cursor;
   const gchar *datasource, *old_version_str;
   gint old_version;
   GList *element;
-  CleanupJob *job = user_data;
-  GomMiner *self = job->self;
-  GomMinerClass *klass = GOM_MINER_GET_CLASS (self);
+  CleanupJob *job;
+  GomMiner *self;
+  GomMinerClass *klass;
+
+  cancellable = g_task_get_cancellable (task);
+  job = (CleanupJob *) g_task_get_task_data (task);
+  self = job->self;
+  klass = GOM_MINER_GET_CLASS (self);
 
   /* find all our datasources in the tracker DB */
   select = g_string_new (NULL);
@@ -568,7 +624,7 @@ cleanup_job (GIOSchedulerJob *sched_job,
 
   cursor = tracker_sparql_connection_query (self->priv->connection,
                                             select->str,
-                                            self->priv->cancellable,
+                                            cancellable,
                                             &error);
   g_string_free (select, TRUE);
 
@@ -578,7 +634,7 @@ cleanup_job (GIOSchedulerJob *sched_job,
       goto out;
     }
 
-  while (tracker_sparql_cursor_next (cursor, self->priv->cancellable, NULL))
+  while (tracker_sparql_cursor_next (cursor, cancellable, NULL))
     {
       /* If the source we found is not in the current list, add
        * it to the cleanup list.
@@ -615,18 +671,22 @@ cleanup_job (GIOSchedulerJob *sched_job,
   g_object_unref (cursor);
 
   /* cleanup the DB */
-  cleanup_job_do_cleanup (job);
+  cleanup_job_do_cleanup (job, cancellable);
 
  out:
-  g_io_scheduler_job_send_to_mainloop_async (sched_job,
-                                             cleanup_old_accounts_done, job, NULL);
-  return FALSE;
+  source = g_idle_source_new ();
+  g_source_set_name (source, "[gnome-online-miners] cleanup_old_accounts_done");
+  g_task_attach_source (task, source, cleanup_old_accounts_done);
+  g_source_unref (source);
+
+  g_object_unref (task);
 }
 
 static void
 gom_miner_cleanup_old_accounts (GomMiner *self,
                                 GList *content_objects,
-                                GList *acc_objects)
+                                GList *acc_objects,
+                                GTask *task)
 {
   CleanupJob *job = g_slice_new0 (CleanupJob);
 
@@ -634,13 +694,12 @@ gom_miner_cleanup_old_accounts (GomMiner *self,
   job->content_objects = content_objects;
   job->acc_objects = acc_objects;
 
-  g_io_scheduler_push_job (cleanup_job, job, NULL,
-                           G_PRIORITY_DEFAULT,
-                           self->priv->cancellable);
+  g_task_set_task_data (task, job, NULL);
+  g_thread_pool_push (cleanup_pool, g_object_ref (task), NULL);
 }
 
 static void
-gom_miner_refresh_db_real (GomMiner *self)
+gom_miner_refresh_db_real (GomMiner *self, GTask *task)
 {
   GoaDocuments *documents;
   GoaPhotos *photos;
@@ -687,26 +746,7 @@ gom_miner_refresh_db_real (GomMiner *self)
 
   g_list_free_full (accounts, g_object_unref);
 
-  gom_miner_cleanup_old_accounts (self, content_objects, acc_objects);
-}
-
-static void
-sparql_connection_ready_cb (GObject *object,
-                            GAsyncResult *res,
-                            gpointer user_data)
-{
-  GError *error = NULL;
-  GomMiner *self = user_data;
-
-  self->priv->connection = tracker_sparql_connection_get_finish (res, &error);
-
-  if (error != NULL)
-    {
-      gom_miner_complete_error (self, error);
-      return;
-    }
-
-  gom_miner_refresh_db_real (self);
+  gom_miner_cleanup_old_accounts (self, content_objects, acc_objects, task);
 }
 
 const gchar *
@@ -715,27 +755,163 @@ gom_miner_get_display_name (GomMiner *self)
   return self->priv->display_name;
 }
 
+static void
+gom_miner_insert_shared_content_in_thread_func (GTask *task,
+                                                gpointer source_object,
+                                                gpointer task_data,
+                                                GCancellable *cancellable)
+{
+  GomMiner *self = GOM_MINER (source_object);
+  GError *error;
+  InsertSharedContentData *data = (InsertSharedContentData *) task_data;
+  gchar *datasource_urn = NULL;
+  gchar *root_element_urn = NULL;
+
+  datasource_urn = g_strdup_printf ("gd:goa-account:%s", data->account_id);
+  root_element_urn = g_strdup_printf ("gd:goa-account:%s:root-element", data->account_id);
+
+  error = NULL;
+  gom_miner_ensure_datasource (self, datasource_urn, root_element_urn, cancellable, &error);
+  if (error != NULL)
+    {
+      g_task_return_error (task, error);
+      goto out;
+    }
+
+  error = NULL;
+  GOM_MINER_GET_CLASS (self)->insert_shared_content (self,
+                                                     data->service,
+                                                     self->priv->connection,
+                                                     datasource_urn,
+                                                     data->shared_id,
+                                                     data->shared_type,
+                                                     data->source_urn,
+                                                     cancellable,
+                                                     &error);
+  if (error != NULL)
+    {
+      g_task_return_error (task, error);
+      goto out;
+    }
+
+  g_task_return_boolean (task, TRUE);
+
+ out:
+  g_free (datasource_urn);
+  g_free (root_element_urn);
+}
+
+void
+gom_miner_insert_shared_content_async (GomMiner *self,
+                                       const gchar *account_id,
+                                       const gchar *shared_id,
+                                       const gchar *shared_type,
+                                       const gchar *source_urn,
+                                       GCancellable *cancellable,
+                                       GAsyncReadyCallback callback,
+                                       gpointer user_data)
+{
+  GTask *task = NULL;
+  GoaDocuments *documents;
+  GoaObject *object = NULL;
+  GoaPhotos *photos;
+  InsertSharedContentData *data;
+  gpointer service;
+
+  task = g_task_new (self, cancellable, callback, user_data);
+  g_task_set_source_tag (task, gom_miner_insert_shared_content_async);
+
+  if (self->priv->client_error != NULL)
+    {
+      g_task_return_error (task, g_error_copy (self->priv->client_error));
+      goto out;
+    }
+
+  if (self->priv->connection_error != NULL)
+    {
+      g_task_return_error (task, g_error_copy (self->priv->connection_error));
+      goto out;
+    }
+
+  object = goa_client_lookup_by_id (self->priv->client, account_id);
+  if (object == NULL)
+    {
+      /* throw error */
+      goto out;
+    }
+
+  documents = goa_object_peek_documents (object);
+  photos = goa_object_peek_photos (object);
+
+  if (g_strcmp0 (shared_type, "documents") == 0 && documents == NULL)
+    {
+      /* throw error */
+      goto out;
+    }
+
+  if (g_strcmp0 (shared_type, "photos") == 0 && photos == NULL)
+    {
+      /* throw error */
+      goto out;
+    }
+
+  service = GOM_MINER_GET_CLASS (self)->create_service (self, object, shared_type);
+  if (service == NULL)
+    {
+      /* throw error */
+      goto out;
+    }
+
+  data = gom_insert_shared_content_data_new (self, account_id, shared_id, shared_type, source_urn, service);
+  g_task_set_task_data (task, data, (GDestroyNotify) gom_insert_shared_content_data_free);
+
+  g_task_run_in_thread (task, gom_miner_insert_shared_content_in_thread_func);
+
+ out:
+  g_clear_object (&object);
+  g_clear_object (&task);
+}
+
+gboolean
+gom_miner_insert_shared_content_finish (GomMiner *self, GAsyncResult *res, GError **error)
+{
+  GTask *task;
+
+  g_assert (g_task_is_valid (res, self));
+  task = G_TASK (res);
+
+  g_assert (g_task_get_source_tag (task) == gom_miner_insert_shared_content_async);
+
+  return g_task_propagate_boolean (task, error);
+}
+
 void
 gom_miner_refresh_db_async (GomMiner *self,
                             GCancellable *cancellable,
                             GAsyncReadyCallback callback,
                             gpointer user_data)
 {
+  GTask *task = NULL;
+
+  task = g_task_new (self, cancellable, callback, user_data);
+  g_task_set_source_tag (task, gom_miner_refresh_db_async);
+
   if (self->priv->client_error != NULL)
     {
-      gom_miner_complete_error (self, self->priv->client_error);
-      return;
+      g_task_return_error (task, g_error_copy (self->priv->client_error));
+      goto out;
     }
 
-  self->priv->result =
-    g_simple_async_result_new (G_OBJECT (self),
-                               callback, user_data,
-                               gom_miner_refresh_db_async);
-  self->priv->cancellable =
-    (cancellable != NULL) ? g_object_ref (cancellable) : NULL;
+  if (self->priv->connection_error != NULL)
+    {
+      g_task_return_error (task, g_error_copy (self->priv->connection_error));
+      goto out;
+    }
 
-  tracker_sparql_connection_get_async (self->priv->cancellable,
-                                       sparql_connection_ready_cb, self);
+  gom_miner_refresh_db_real (self, task);
+
+ out:
+  g_clear_object (&task);
 }
 
 gboolean
@@ -743,15 +919,14 @@ gom_miner_refresh_db_finish (GomMiner *self,
                              GAsyncResult *res,
                              GError **error)
 {
-  GSimpleAsyncResult *simple_res = G_SIMPLE_ASYNC_RESULT (res);
+  GTask *task;
 
-  g_assert (g_simple_async_result_is_valid (res, G_OBJECT (self),
-                                            gom_miner_refresh_db_async));
+  g_assert (g_task_is_valid (res, self));
+  task = G_TASK (res);
 
-  if (g_simple_async_result_propagate_error (simple_res, error))
-    return FALSE;
+  g_assert (g_task_get_source_tag (task) == gom_miner_refresh_db_async);
 
-  return TRUE;
+  return g_task_propagate_boolean (task, error);
 }
 
 void
